@@ -48,6 +48,14 @@ extern DLLIMPORT char *Log_directory;
 
 #include "pgut/pgut-be.h"
 
+#if PG_VERSION_NUM >= 90100
+#define RelationIsTemp(rel)	\
+	((rel)->rd_rel->relpersistence == RELPERSISTENCE_TEMP)
+#else
+#define RelationIsTemp(rel)	 \
+	((rel)->rd_istemp)
+#endif
+
 PG_MODULE_MAGIC;
 
 /*
@@ -181,11 +189,11 @@ extern Datum PGUT_EXPORT senna_costestimate(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT senna_options(PG_FUNCTION_ARGS);
 #endif
 
-static char *SennaPath(const RelFileNode *rnode);
+static char *SennaPath(const RelFileNode *rnode, bool istemp);
 static sen_encoding SennaEncoding(void);
 static sen_index *SennaIndexOpen(Relation index, int flags);
 static void SennaIndexClose(Oid relid, const RelFileNode *rnode);
-static bool SennaRemove(const RelFileNode *rnode);
+static bool SennaRemove(const RelFileNode *rnode, bool istemp);
 static sen_query *SennaQuery(const char *key, size_t len);
 static sen_records *SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **idx);
 static void SennaRecordsClose(sen_records *records);
@@ -354,6 +362,7 @@ senna_drop_index(PG_FUNCTION_ARGS)
 	Oid			relid = PG_GETARG_OID(0);
 	Relation	index;
 	RelFileNode	rnode;
+	bool		istemp;
 	DropStmt   *stmt;
 	List	   *obj;
 
@@ -370,11 +379,12 @@ senna_drop_index(PG_FUNCTION_ARGS)
 	stmt->behavior = DROP_CASCADE;
 
 	rnode = index->rd_node;
+	istemp = RelationIsTemp(index);
 	index_close(index, NoLock);
 
 	RemoveRelations(stmt);
 
-	SennaRemove(&rnode);
+	SennaRemove(&rnode, istemp);
 
 	PG_RETURN_VOID();
 }
@@ -385,14 +395,16 @@ senna_reindex_index(PG_FUNCTION_ARGS)
 	Oid			relid = PG_GETARG_OID(0);
 	Relation	index;
 	RelFileNode	rnode;
+	bool		istemp;
 
 	index = senna_index_open(relid, NoLock);
 	rnode = index->rd_node;
+	istemp = RelationIsTemp(index);
 	index_close(index, NoLock);
 
 	reindex_index(relid, false);
 
-	SennaRemove(&rnode);
+	SennaRemove(&rnode, istemp);
 
 	PG_RETURN_VOID();
 }
@@ -561,16 +573,27 @@ Datum
 senna_beginscan(PG_FUNCTION_ARGS)
 {
 	Relation		index = (Relation) PG_GETARG_POINTER(0);
-	int				keysz = PG_GETARG_INT32(1);
+	int				nkeys = PG_GETARG_INT32(1);
+#if PG_VERSION_NUM >= 90100
+	int			norderbys = PG_GETARG_INT32(2);
+#else
 	ScanKey			key = (ScanKey) PG_GETARG_POINTER(2);
+#endif
 	IndexScanDesc	scan;
 
-	if (keysz < 1)
+	if (nkeys < 1)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("senna: do not support whole-index scans")));
 
-	scan = RelationGetIndexScan(index, keysz, key);
+#if PG_VERSION_NUM >= 90100
+	/* no order by operators allowed */
+	Assert(norderbys == 0);
+
+	scan = RelationGetIndexScan(index, nkeys, norderbys);
+#else
+	scan = RelationGetIndexScan(index, nkeys, key);
+#endif
 
 	PG_RETURN_POINTER(scan);
 }
@@ -710,7 +733,7 @@ senna_getmulti(PG_FUNCTION_ARGS)
 		tids[ntids] = ctid;
 	}
 
-	*returned_tids = ntids;
+	*returned_tids = (int32) ntids;
 	PG_RETURN_BOOL(ntids >= max_tids);
 }
 
@@ -779,6 +802,7 @@ senna_build(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("senna: do not support multi columns")));
 
+	/* FIXME: Relation->rd_operator is missing in 9.1 :-( */
 	if (OPCLASS_IS_NORM(index->rd_operator))
 		flags = FLAGS_NORM;
 	else if (OPCLASS_IS_LIKE(index->rd_operator))
@@ -816,7 +840,7 @@ senna_build(PG_FUNCTION_ARGS)
 	}
 	PG_CATCH();
 	{
-		SennaRemove(&index->rd_node);
+		SennaRemove(&index->rd_node, RelationIsTemp(index));
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -1009,9 +1033,16 @@ senna_options(PG_FUNCTION_ARGS)
 #endif
 
 static char *
-SennaPath(const RelFileNode *rnode)
+SennaPath(const RelFileNode *rnode, bool istemp)
 {
+#if PG_VERSION_NUM >= 90100
+	if (istemp)
+		return relpathbackend(*rnode, MyBackendId, MAIN_FORKNUM);
+	else
+		return relpathbackend(*rnode, InvalidBackendId, MAIN_FORKNUM);
+#else
 	return relpath(*rnode, MAIN_FORKNUM);
+#endif
 }
 
 static sen_encoding	sen_encoding_cache = sen_enc_default;
@@ -1077,7 +1108,7 @@ SennaIndexOpen(Relation indexRelation, int flags)
 	}
 
 	/* Be careful not to leak resources on error. */
-	path = SennaPath(&indexRelation->rd_node);
+	path = SennaPath(&indexRelation->rd_node, RelationIsTemp(indexRelation));
 	if (flags != 0)
 	{
 		DEBUG_INDEX_CREATE(indexRelation->rd_node);
@@ -1161,14 +1192,14 @@ SennaIndexClose(Oid relid, const RelFileNode *rnode)
 }
 
 static bool
-SennaRemove(const RelFileNode *rnode)
+SennaRemove(const RelFileNode *rnode, bool istemp)
 {
 	int			loops = 0;
 	sen_rc		rc;
 	char	   *path;
 
 	SennaIndexClose(InvalidOid, rnode);
-	path = SennaPath(rnode);
+	path = SennaPath(rnode, istemp);
 
 	while ((rc = sen_index_remove(path)) != sen_success)
 	{
